@@ -1,9 +1,9 @@
 // Copyright © Tavian Barnes <tavianator@tavianator.com>
 // SPDX-License-Identifier: 0BSD
 
+#include "prelude.h"
 #include "alloc.h"
 #include "bit.h"
-#include "config.h"
 #include "diag.h"
 #include "sanity.h"
 #include <errno.h>
@@ -24,12 +24,12 @@ static void *xmemalign(size_t align, size_t size) {
 	bfs_assert(align >= sizeof(void *));
 	bfs_assert(is_aligned(align, size));
 
-#if __APPLE__
+#if BFS_HAS_ALIGNED_ALLOC
+	return aligned_alloc(align, size);
+#else
 	void *ptr = NULL;
 	errno = posix_memalign(&ptr, align, size);
 	return ptr;
-#else
-	return aligned_alloc(align, size);
 #endif
 }
 
