@@ -1,12 +1,14 @@
 // Copyright © Tavian Barnes <tavianator@tavianator.com>
 // SPDX-License-Identifier: 0BSD
 
+#include "prelude.h"
 #include "bar.h"
+#include "alloc.h"
 #include "atomic.h"
 #include "bfstd.h"
 #include "bit.h"
-#include "config.h"
 #include "dstring.h"
+#include "sighook.h"
 #include <errno.h>
 #include <fcntl.h>
 #include <signal.h>
@@ -19,11 +21,9 @@ struct bfs_bar {
 	int fd;
 	atomic unsigned int width;
 	atomic unsigned int height;
-};
 
-/** The global status bar instance. */
-static struct bfs_bar the_bar = {
-	.fd = -1,
+	struct sighook *exit_hook;
+	struct sighook *winch_hook;
 };
 
 /** Get the terminal size, if possible. */
@@ -43,10 +43,14 @@ static int bfs_bar_getsize(struct bfs_bar *bar) {
 #endif
 }
 
-/** Async Signal Safe puts(). */
-static int ass_puts(int fd, const char *str) {
-	size_t len = strlen(str);
-	return xwrite(fd, str, len) == len ? 0 : -1;
+/** Write a string to the status bar (async-signal-safe). */
+static int bfs_bar_write(struct bfs_bar *bar, const char *str, size_t len) {
+	return xwrite(bar->fd, str, len) == len ? 0 : -1;
+}
+
+/** Write a string to the status bar (async-signal-safe). */
+static int bfs_bar_puts(struct bfs_bar *bar, const char *str) {
+	return bfs_bar_write(bar, str, strlen(str));
 }
 
 /** Number of decimal digits needed for terminal sizes. */
@@ -70,39 +74,40 @@ static char *ass_itoa(char *str, unsigned int n) {
 
 /** Update the size of the scrollable region. */
 static int bfs_bar_resize(struct bfs_bar *bar) {
-	char esc_seq[12 + ITOA_DIGITS] =
+	static const char PREFIX[] =
+		"\033D"   // IND: Line feed, possibly scrolling
+		"\033[1A" // CUU: Move cursor up 1 row
 		"\0337"   // DECSC: Save cursor
 		"\033[;"; // DECSTBM: Set scrollable region
+	static const char SUFFIX[] =
+		"r"       // (end of DECSTBM)
+		"\0338"   // DECRC: Restore the cursor
+		"\033[J"; // ED: Erase display from cursor to end
+
+	char esc_seq[sizeof(PREFIX) + ITOA_DIGITS + sizeof(SUFFIX)];
 
 	// DECSTBM takes the height as the second argument
-	unsigned int height = load(&bar->height, relaxed);
-	char *ptr = esc_seq + strlen(esc_seq);
-	ptr = ass_itoa(ptr, height - 1);
+	unsigned int height = load(&bar->height, relaxed) - 1;
 
-	strcpy(ptr,
-		"r"      // DECSTBM
-		"\0338"  // DECRC: Restore the cursor
-		"\033[J" // ED: Erase display from cursor to end
-	);
+	char *cur = stpcpy(esc_seq, PREFIX);
+	cur = ass_itoa(cur, height);
+	cur = stpcpy(cur, SUFFIX);
 
-	return ass_puts(bar->fd, esc_seq);
+	return bfs_bar_write(bar, esc_seq, cur - esc_seq);
 }
 
 #ifdef SIGWINCH
 /** SIGWINCH handler. */
-static void sighand_winch(int sig) {
-	int error = errno;
-
-	bfs_bar_getsize(&the_bar);
-	bfs_bar_resize(&the_bar);
-
-	errno = error;
+static void bfs_bar_sigwinch(int sig, siginfo_t *info, void *arg) {
+	struct bfs_bar *bar = arg;
+	bfs_bar_getsize(bar);
+	bfs_bar_resize(bar);
 }
 #endif
 
 /** Reset the scrollable region and hide the bar. */
 static int bfs_bar_reset(struct bfs_bar *bar) {
-	return ass_puts(bar->fd,
+	return bfs_bar_puts(bar,
 		"\0337"  // DECSC: Save cursor
 		"\033[r" // DECSTBM: Reset scrollable region
 		"\0338"  // DECRC: Restore cursor
@@ -111,19 +116,9 @@ static int bfs_bar_reset(struct bfs_bar *bar) {
 }
 
 /** Signal handler for process-terminating signals. */
-static void sighand_reset(int sig) {
-	bfs_bar_reset(&the_bar);
-	raise(sig);
-}
-
-/** Register sighand_reset() for a signal. */
-static void reset_before_death_by(int sig) {
-	struct sigaction sa = {
-		.sa_handler = sighand_reset,
-		.sa_flags = SA_RESETHAND,
-	};
-	sigemptyset(&sa.sa_mask);
-	sigaction(sig, &sa, NULL);
+static void bfs_bar_sigexit(int sig, siginfo_t *info, void *arg) {
+	struct bfs_bar *bar = arg;
+	bfs_bar_reset(bar);
 }
 
 /** printf() to the status bar with a single write(). */
@@ -138,64 +133,47 @@ static int bfs_bar_printf(struct bfs_bar *bar, const char *format, ...) {
 		return -1;
 	}
 
-	int ret = ass_puts(bar->fd, str);
+	int ret = bfs_bar_write(bar, str, dstrlen(str));
 	dstrfree(str);
 	return ret;
 }
 
 struct bfs_bar *bfs_bar_show(void) {
-	if (the_bar.fd >= 0) {
-		errno = EBUSY;
+	struct bfs_bar *bar = ALLOC(struct bfs_bar);
+	if (!bar) {
+		return NULL;
+	}
+
+	bar->fd = open_cterm(O_RDWR | O_CLOEXEC);
+	if (bar->fd < 0) {
 		goto fail;
 	}
 
-	char term[L_ctermid];
-	ctermid(term);
-	if (strlen(term) == 0) {
-		errno = ENOTTY;
-		goto fail;
-	}
-
-	the_bar.fd = open(term, O_RDWR | O_CLOEXEC);
-	if (the_bar.fd < 0) {
-		goto fail;
-	}
-
-	if (bfs_bar_getsize(&the_bar) != 0) {
+	if (bfs_bar_getsize(bar) != 0) {
 		goto fail_close;
 	}
 
-	reset_before_death_by(SIGABRT);
-	reset_before_death_by(SIGINT);
-	reset_before_death_by(SIGPIPE);
-	reset_before_death_by(SIGQUIT);
-	reset_before_death_by(SIGTERM);
+	bar->exit_hook = atsigexit(bfs_bar_sigexit, bar);
+	if (!bar->exit_hook) {
+		goto fail_close;
+	}
 
 #ifdef SIGWINCH
-	struct sigaction sa = {
-		.sa_handler = sighand_winch,
-		.sa_flags = SA_RESTART,
-	};
-	sigemptyset(&sa.sa_mask);
-	sigaction(SIGWINCH, &sa, NULL);
+	bar->winch_hook = sighook(SIGWINCH, bfs_bar_sigwinch, bar, 0);
+	if (!bar->winch_hook) {
+		goto fail_hook;
+	}
 #endif
 
-	unsigned int height = load(&the_bar.height, relaxed);
-	bfs_bar_printf(&the_bar,
-		"\n"        // Make space for the bar
-		"\0337"     // DECSC: Save cursor
-		"\033[;%ur" // DECSTBM: Set scrollable region
-		"\0338"     // DECRC: Restore cursor
-		"\033[1A",  // CUU: Move cursor up 1 row
-		height - 1
-	);
-
-	return &the_bar;
+	bfs_bar_resize(bar);
+	return bar;
 
+fail_hook:
+	sigunhook(bar->exit_hook);
 fail_close:
-	close_quietly(the_bar.fd);
-	the_bar.fd = -1;
+	close_quietly(bar->fd);
 fail:
+	free(bar);
 	return NULL;
 }
 
@@ -223,17 +201,11 @@ void bfs_bar_hide(struct bfs_bar *bar) {
 		return;
 	}
 
-	signal(SIGABRT, SIG_DFL);
-	signal(SIGINT, SIG_DFL);
-	signal(SIGPIPE, SIG_DFL);
-	signal(SIGQUIT, SIG_DFL);
-	signal(SIGTERM, SIG_DFL);
-#ifdef SIGWINCH
-	signal(SIGWINCH, SIG_DFL);
-#endif
+	sigunhook(bar->winch_hook);
+	sigunhook(bar->exit_hook);
 
 	bfs_bar_reset(bar);
 
 	xclose(bar->fd);
-	bar->fd = -1;
+	free(bar);
 }
