@@ -8,7 +8,7 @@
 #ifndef BFS_BFSTD_H
 #define BFS_BFSTD_H
 
-#include "config.h"
+#include "prelude.h"
 #include "sanity.h"
 #include <stddef.h>
 
@@ -158,6 +158,16 @@ FILE *xfopen(const char *path, int flags);
  */
 char *xgetdelim(FILE *file, char delim);
 
+/**
+ * Open the controlling terminal.
+ *
+ * @param flags
+ *         The open() flags.
+ * @return
+ *         An open file descriptor, or -1 on failure.
+ */
+int open_cterm(int flags);
+
 // #include <stdlib.h>
 
 /**
@@ -177,6 +187,21 @@ int ynprompt(void);
 
 // #include <string.h>
 
+/**
+ * Get the length of the pure-ASCII prefix of a string.
+ */
+size_t asciilen(const char *str);
+
+/**
+ * Get the length of the pure-ASCII prefix of a string.
+ *
+ * @param str
+ *         The string to check.
+ * @param n
+ *         The maximum prefix length.
+ */
+size_t asciinlen(const char *str, size_t n);
+
 /**
  * Allocate a copy of a region of memory.
  *
@@ -232,6 +257,11 @@ char *xstpencpy(char *dest, char *end, const char *src, size_t n);
  */
 const char *xstrerror(int errnum);
 
+/**
+ * Shorthand for xstrerror(errno).
+ */
+const char *errstr(void);
+
 /**
  * Format a mode like ls -l (e.g. -rw-r--r--).
  *
@@ -268,11 +298,21 @@ int xminor(dev_t dev);
 
 // #include <sys/stat.h>
 
-#if __APPLE__
-#  define st_atim st_atimespec
-#  define st_ctim st_ctimespec
-#  define st_mtim st_mtimespec
-#  define st_birthtim st_birthtimespec
+/**
+ * Get the access/change/modification time from a struct stat.
+ */
+#if BFS_HAS_ST_ACMTIM
+#  define ST_ATIM(sb) (sb).st_atim
+#  define ST_CTIM(sb) (sb).st_ctim
+#  define ST_MTIM(sb) (sb).st_mtim
+#elif BFS_HAS_ST_ACMTIMESPEC
+#  define ST_ATIM(sb) (sb).st_atimespec
+#  define ST_CTIM(sb) (sb).st_ctimespec
+#  define ST_MTIM(sb) (sb).st_mtimespec
+#else
+#  define ST_ATIM(sb) ((struct timespec) { .tv_sec = (sb).st_atime })
+#  define ST_CTIM(sb) ((struct timespec) { .tv_sec = (sb).st_ctime })
+#  define ST_MTIM(sb) ((struct timespec) { .tv_sec = (sb).st_mtime })
 #endif
 
 // #include <sys/wait.h>
@@ -385,6 +425,11 @@ char *xconfstr(int name);
  */
 int xstrtofflags(const char **str, unsigned long long *set, unsigned long long *clear);
 
+/**
+ * Wrapper for sysconf() that works around an MSan bug.
+ */
+long xsysconf(int name);
+
 #include <wchar.h>
 
 /**
