@@ -8,9 +8,8 @@
 #ifndef BFS_COLOR_H
 #define BFS_COLOR_H
 
-#include "config.h"
+#include "prelude.h"
 #include "dstring.h"
-#include <stdarg.h>
 #include <stdio.h>
 
 /**
@@ -43,6 +42,8 @@ typedef struct CFILE {
 	const struct colors *colors;
 	/** A buffer for colored formatting. */
 	dchar *buffer;
+	/** Cached file descriptor number. */
+	int fd;
 	/** Whether the next ${rs} is actually necessary. */
 	bool need_reset;
 	/** Whether to close the underlying stream. */
@@ -86,7 +87,6 @@ int cfclose(CFILE *cfile);
  *         %g: A double
  *         %s: A string
  *         %zu: A size_t
- *         %m: strerror(errno)
  *         %pq: A shell-escaped string, like bash's printf %q
  *         %pQ: A TTY-escaped string.
  *         %pF: A colored file name, from a const struct BFTW * argument
@@ -109,4 +109,9 @@ int cfprintf(CFILE *cfile, const char *format, ...);
 attr(printf(2, 0))
 int cvfprintf(CFILE *cfile, const char *format, va_list args);
 
+/**
+ * Reset the TTY state when terminating abnormally (async-signal-safe).
+ */
+int cfreset(CFILE *cfile);
+
 #endif // BFS_COLOR_H
