@@ -1,18 +1,18 @@
 // Copyright © Tavian Barnes <tavianator@tavianator.com>
 // SPDX-License-Identifier: 0BSD
 
+#include "prelude.h"
 #include "diag.h"
 #include "alloc.h"
 #include "bfstd.h"
 #include "color.h"
-#include "config.h"
 #include "ctx.h"
 #include "dstring.h"
 #include "expr.h"
 #include <errno.h>
 #include <stdarg.h>
+#include <stdio.h>
 #include <stdlib.h>
-#include <string.h>
 
 /** bfs_diagf() implementation. */
 attr(printf(2, 0))
@@ -64,7 +64,7 @@ const char *debug_flag_name(enum debug_flags flag) {
 }
 
 void bfs_perror(const struct bfs_ctx *ctx, const char *str) {
-	bfs_error(ctx, "%s: %m.\n", str);
+	bfs_error(ctx, "%s: %s.\n", str, errstr());
 }
 
 void bfs_error(const struct bfs_ctx *ctx, const char *format, ...) {
@@ -91,19 +91,12 @@ bool bfs_debug(const struct bfs_ctx *ctx, enum debug_flags flag, const char *for
 }
 
 void bfs_verror(const struct bfs_ctx *ctx, const char *format, va_list args) {
-	int error = errno;
-
 	bfs_error_prefix(ctx);
-
-	errno = error;
 	cvfprintf(ctx->cerr, format, args);
 }
 
 bool bfs_vwarning(const struct bfs_ctx *ctx, const char *format, va_list args) {
-	int error = errno;
-
 	if (bfs_warning_prefix(ctx)) {
-		errno = error;
 		cvfprintf(ctx->cerr, format, args);
 		return true;
 	} else {
@@ -112,10 +105,7 @@ bool bfs_vwarning(const struct bfs_ctx *ctx, const char *format, va_list args) {
 }
 
 bool bfs_vdebug(const struct bfs_ctx *ctx, enum debug_flags flag, const char *format, va_list args) {
-	int error = errno;
-
 	if (bfs_debug_prefix(ctx, flag)) {
-		errno = error;
 		cvfprintf(ctx->cerr, format, args);
 		return true;
 	} else {
@@ -169,7 +159,7 @@ static bool highlight_expr_recursive(const struct bfs_ctx *ctx, const struct bfs
 		}
 	}
 
-	for (struct bfs_expr *child = bfs_expr_children(expr); child; child = child->next) {
+	for_expr (child, expr) {
 		ret |= highlight_expr_recursive(ctx, child, args);
 	}
 
