@@ -8,8 +8,7 @@
 #ifndef BFS_DIR_H
 #define BFS_DIR_H
 
-#include "alloc.h"
-#include "config.h"
+#include "prelude.h"
 #include <sys/types.h>
 
 /**
@@ -17,7 +16,11 @@
  * libc's readdir().
  */
 #ifndef BFS_USE_GETDENTS
-#  define BFS_USE_GETDENTS (__linux__ || __FreeBSD__)
+#  if BFS_HAS_POSIX_GETDENTS
+#    define BFS_USE_GETDENTS true
+#  elif __linux__ || __FreeBSD__
+#    define BFS_USE_GETDENTS (BFS_HAS_GETDENTS || BFS_HAS_GETDENTS64 | BFS_HAS_GETDENTS64_SYSCALL)
+#  endif
 #endif
 
 /**
@@ -78,6 +81,8 @@ struct bfs_dirent {
  */
 struct bfs_dir *bfs_allocdir(void);
 
+struct arena;
+
 /**
  * Initialize an arena for directories.
  *
@@ -152,7 +157,7 @@ int bfs_closedir(struct bfs_dir *dir);
  * Whether the bfs_unwrapdir() function is supported.
  */
 #ifndef BFS_USE_UNWRAPDIR
-#  define BFS_USE_UNWRAPDIR (BFS_USE_GETDENTS || __FreeBSD__)
+#  define BFS_USE_UNWRAPDIR (BFS_USE_GETDENTS || BFS_HAS_FDCLOSEDIR)
 #endif
 
 #if BFS_USE_UNWRAPDIR
