@@ -25,12 +25,14 @@
  * effects are reachable at all, skipping the traversal if not.
  */
 
+#include "prelude.h"
 #include "opt.h"
+#include "bftw.h"
 #include "bit.h"
 #include "color.h"
-#include "config.h"
 #include "ctx.h"
 #include "diag.h"
+#include "dir.h"
 #include "eval.h"
 #include "exec.h"
 #include "expr.h"
@@ -40,7 +42,6 @@
 #include <limits.h>
 #include <stdarg.h>
 #include <stdio.h>
-#include <string.h>
 #include <unistd.h>
 
 static char *fake_and_arg = "-and";
@@ -723,7 +724,7 @@ struct visitor {
 	visit_fn *leave;
 
 	/** A visitor lookup table. */
-	struct visitor_table table[];
+	const struct visitor_table *table;
 };
 
 /** Recursive visitor implementation. */
@@ -946,8 +947,12 @@ static struct bfs_expr *visit_shallow(struct bfs_opt *opt, struct bfs_expr *expr
 		expr = general(opt, expr, visitor);
 	}
 
+	if (!expr) {
+		return NULL;
+	}
+
 	visit_fn *specific = look_up_visitor(expr, visitor->table);
-	if (expr && specific) {
+	if (specific) {
 		expr = specific(opt, expr, visitor);
 	}
 
@@ -1083,7 +1088,7 @@ static struct bfs_expr *annotate_and(struct bfs_opt *opt, struct bfs_expr *expr,
 	expr->cost = 0.0;
 	expr->probability = 1.0;
 
-	for (struct bfs_expr *child = bfs_expr_children(expr); child; child = child->next) {
+	for_expr (child, expr) {
 		expr->pure &= child->pure;
 		expr->always_true &= child->always_true;
 		expr->always_false |= child->always_false;
@@ -1102,7 +1107,7 @@ static struct bfs_expr *annotate_or(struct bfs_opt *opt, struct bfs_expr *expr,
 	expr->cost = 0.0;
 
 	float false_prob = 1.0;
-	for (struct bfs_expr *child = bfs_expr_children(expr); child; child = child->next) {
+	for_expr (child, expr) {
 		expr->pure &= child->pure;
 		expr->always_true |= child->always_true;
 		expr->always_false &= child->always_false;
@@ -1119,7 +1124,7 @@ static struct bfs_expr *annotate_comma(struct bfs_opt *opt, struct bfs_expr *exp
 	expr->pure = true;
 	expr->cost = 0.0;
 
-	for (struct bfs_expr *child = bfs_expr_children(expr); child; child = child->next) {
+	for_expr (child, expr) {
 		expr->pure &= child->pure;
 		expr->always_true = child->always_true;
 		expr->always_false = child->always_false;
@@ -1180,6 +1185,7 @@ static struct bfs_expr *annotate_visit(struct bfs_opt *opt, struct bfs_expr *exp
 		eval_fprint0,
 		eval_fprintf,
 		eval_fprintx,
+		eval_limit,
 		eval_prune,
 		eval_true,
 		// Non-returning
@@ -1329,7 +1335,7 @@ static struct bfs_expr *annotate_visit(struct bfs_opt *opt, struct bfs_expr *exp
 static const struct visitor annotate = {
 	.name = "annotate",
 	.visit = annotate_visit,
-	.table = {
+	.table = (const struct visitor_table[]) {
 		{eval_access, annotate_access},
 		{eval_empty, annotate_empty},
 		{eval_exec, annotate_exec},
@@ -1514,7 +1520,7 @@ static struct bfs_expr *canonicalize_assoc(struct bfs_opt *opt, struct bfs_expr
  */
 static const struct visitor canonicalize = {
 	.name = "canonicalize",
-	.table = {
+	.table = (const struct visitor_table[]) {
 		{eval_not, canonicalize_not},
 		{eval_and, canonicalize_assoc},
 		{eval_or, canonicalize_assoc},
@@ -1607,7 +1613,7 @@ static struct bfs_expr *reorder_andor(struct bfs_opt *opt, struct bfs_expr *expr
  */
 static const struct visitor reorder = {
 	.name = "reorder",
-	.table = {
+	.table = (const struct visitor_table[]) {
 		{eval_and, reorder_andor},
 		{eval_or, reorder_andor},
 		{NULL, NULL},
@@ -1784,7 +1790,7 @@ static struct bfs_expr *data_flow_leave(struct bfs_opt *opt, struct bfs_expr *ex
 	if (df_is_bottom(&opt->after_false)) {
 		if (!expr->pure) {
 			expr->always_true = true;
-			expr->probability = 0.0;
+			expr->probability = 1.0;
 		} else if (expr->eval_fn != eval_true) {
 			opt_warning(opt, expr, "This expression is always true.\n\n");
 			opt_debug(opt, "pure, always true\n");
@@ -1885,7 +1891,7 @@ static const struct visitor data_flow = {
 	.enter = data_flow_enter,
 	.visit = data_flow_visit,
 	.leave = data_flow_leave,
-	.table = {
+	.table = (const struct visitor_table[]) {
 		{eval_access, data_flow_access},
 		{eval_gid, data_flow_gid},
 		{eval_inum, data_flow_inum},
@@ -1913,7 +1919,7 @@ static struct bfs_expr *simplify_not(struct bfs_opt *opt, struct bfs_expr *expr,
 static struct bfs_expr *lift_andor_not(struct bfs_opt *opt, struct bfs_expr *expr) {
 	// Only lift negations if it would reduce the number of (-not) expressions
 	size_t added = 0, removed = 0;
-	for (struct bfs_expr *child = bfs_expr_children(expr); child; child = child->next) {
+	for_expr (child, expr) {
 		if (child->eval_fn == eval_not) {
 			++removed;
 		} else {
@@ -1962,7 +1968,7 @@ static struct bfs_expr *first_ignorable(struct bfs_opt *opt, struct bfs_expr *ex
 	}
 
 	struct bfs_expr *ret = NULL;
-	for (struct bfs_expr *child = bfs_expr_children(expr); child; child = child->next) {
+	for_expr (child, expr) {
 		if (!child->pure) {
 			ret = NULL;
 		} else if (!ret) {
@@ -2101,7 +2107,7 @@ static struct bfs_expr *simplify_comma(struct bfs_opt *opt, struct bfs_expr *exp
  */
 static const struct visitor simplify = {
 	.name = "simplify",
-	.table = {
+	.table = (const struct visitor_table[]) {
 		{eval_not, simplify_not},
 		{eval_and, simplify_and},
 		{eval_or, simplify_or},
@@ -2170,17 +2176,20 @@ static struct bfs_expr *optimize(struct bfs_opt *opt, struct bfs_expr *expr) {
 	return expr;
 }
 
-/** Estimate the odds of an expression calling stat(). */
-static float expr_stat_odds(struct bfs_expr *expr) {
-	if (expr->calls_stat) {
+/** An expression predicate. */
+typedef bool expr_pred(const struct bfs_expr *expr);
+
+/** Estimate the odds that a matching expression will be evaluated. */
+static float estimate_odds(const struct bfs_expr *expr, expr_pred *pred) {
+	if (pred(expr)) {
 		return 1.0;
 	}
 
-	float nostat_odds = 1.0;
+	float nonmatch_odds = 1.0;
 	float reached_odds = 1.0;
-	for (struct bfs_expr *child = bfs_expr_children(expr); child; child = child->next) {
-		float child_odds = expr_stat_odds(child);
-		nostat_odds *= 1.0 - reached_odds * child_odds;
+	for_expr (child, expr) {
+		float child_odds = estimate_odds(child, pred);
+		nonmatch_odds *= 1.0 - reached_odds * child_odds;
 
 		if (expr->eval_fn == eval_and) {
 			reached_odds *= child->probability;
@@ -2189,7 +2198,12 @@ static float expr_stat_odds(struct bfs_expr *expr) {
 		}
 	}
 
-	return 1.0 - nostat_odds;
+	return 1.0 - nonmatch_odds;
+}
+
+/** Whether an expression calls stat(). */
+static bool calls_stat(const struct bfs_expr *expr) {
+	return expr->calls_stat;
 }
 
 /** Estimate the odds of calling stat(). */
@@ -2198,15 +2212,20 @@ static float estimate_stat_odds(struct bfs_ctx *ctx) {
 		return 1.0;
 	}
 
-	float nostat_odds = 1.0 - expr_stat_odds(ctx->exclude);
+	float nostat_odds = 1.0 - estimate_odds(ctx->exclude, calls_stat);
 
 	float reached_odds = 1.0 - ctx->exclude->probability;
-	float expr_odds = expr_stat_odds(ctx->expr);
+	float expr_odds = estimate_odds(ctx->expr, calls_stat);
 	nostat_odds *= 1.0 - reached_odds * expr_odds;
 
 	return 1.0 - nostat_odds;
 }
 
+/** Matches -(exec|ok) ... \; */
+static bool single_exec(const struct bfs_expr *expr) {
+	return expr->eval_fn == eval_exec && !(expr->exec->flags & BFS_EXEC_MULTI);
+}
+
 int bfs_optimize(struct bfs_ctx *ctx) {
 	bfs_ctx_dump(ctx, DEBUG_OPT);
 
@@ -2285,6 +2304,17 @@ int bfs_optimize(struct bfs_ctx *ctx) {
 			opt_leave(&opt, "eager stat cost: ${ylw}%g${rs}\n", eager_cost);
 		}
 
+#ifndef POSIX_SPAWN_SETRLIMIT
+		// If bfs_spawn_setrlimit() would force us to use fork() over
+		// posix_spawn(), the extra cost may outweigh the benefit of a
+		// higher RLIMIT_NOFILE
+		float single_exec_odds = estimate_odds(ctx->expr, single_exec);
+		if (single_exec_odds >= 0.5) {
+			opt_enter(&opt, "single ${blu}-exec${rs} odds: ${ylw}%g${rs}\n", single_exec_odds);
+			ctx->raise_nofile = false;
+			opt_leave(&opt, "not raising RLIMIT_NOFILE\n");
+		}
+#endif
 	}
 
 	opt_leave(&opt, NULL);
