@@ -15,8 +15,12 @@ struct bfs_regex;
 enum bfs_regex_type {
 	BFS_REGEX_POSIX_BASIC,
 	BFS_REGEX_POSIX_EXTENDED,
+	BFS_REGEX_AWK,
+	BFS_REGEX_GNU_AWK,
 	BFS_REGEX_EMACS,
 	BFS_REGEX_GREP,
+	BFS_REGEX_EGREP,
+	BFS_REGEX_GNU_FIND,
 };
 
 /**
