@@ -8,7 +8,7 @@
 #ifndef BFS_XSPAWN_H
 #define BFS_XSPAWN_H
 
-#include "config.h"
+#include "prelude.h"
 #include <sys/resource.h>
 #include <sys/types.h>
 #include <unistd.h>
@@ -61,6 +61,14 @@ int bfs_spawn_init(struct bfs_spawn *ctx);
  */
 int bfs_spawn_destroy(struct bfs_spawn *ctx);
 
+/**
+ * Add an open() action to a bfs_spawn() context.
+ *
+ * @return
+ *         0 on success, -1 on failure.
+ */
+int bfs_spawn_addopen(struct bfs_spawn *ctx, int fd, const char *path, int flags, mode_t mode);
+
 /**
  * Add a close() action to a bfs_spawn() context.
  *
