@@ -1,15 +1,12 @@
 // Copyright © Tavian Barnes <tavianator@tavianator.com>
 // SPDX-License-Identifier: 0BSD
 
+#include "prelude.h"
 #include "tests.h"
-#include "../src/bfstd.h"
-#include "../src/config.h"
-#include "../src/diag.h"
+#include "bfstd.h"
+#include "diag.h"
 #include <errno.h>
 #include <langinfo.h>
-#include <locale.h>
-#include <stdint.h>
-#include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 
@@ -18,12 +15,12 @@ static bool check_base_dir(const char *path, const char *dir, const char *base)
 	bool ret = true;
 
 	char *xdir = xdirname(path);
-	bfs_verify(xdir, "xdirname(): %s", xstrerror(errno));
+	bfs_everify(xdir, "xdirname()");
 	ret &= bfs_check(strcmp(xdir, dir) == 0, "xdirname('%s') == '%s' (!= '%s')", path, xdir, dir);
 	free(xdir);
 
 	char *xbase = xbasename(path);
-	bfs_verify(xbase, "xbasename(): %s", xstrerror(errno));
+	bfs_everify(xbase, "xbasename()");
 	ret &= bfs_check(strcmp(xbase, base) == 0, "xbasename('%s') == '%s' (!= '%s')", path, xbase, base);
 	free(xbase);
 
@@ -43,10 +40,20 @@ static bool check_wordesc(const char *str, const char *exp, enum wesc_flags flag
 bool check_bfstd(void) {
 	bool ret = true;
 
-	// Try to set a UTF-8 locale
-	if (!setlocale(LC_ALL, "C.UTF-8")) {
-		setlocale(LC_ALL, "");
-	}
+	ret &= bfs_check(asciilen("") == 0);
+	ret &= bfs_check(asciilen("@") == 1);
+	ret &= bfs_check(asciilen("@@") == 2);
+	ret &= bfs_check(asciilen("\xFF@") == 0);
+	ret &= bfs_check(asciilen("@\xFF") == 1);
+	ret &= bfs_check(asciilen("@@@@@@@@") == 8);
+	ret &= bfs_check(asciilen("@@@@@@@@@@@@@@@@") == 16);
+	ret &= bfs_check(asciilen("@@@@@@@@@@@@@@@@@@@@@@@@") == 24);
+	ret &= bfs_check(asciilen("@@@@@@@@@@@@@@a\xFF@@@@@@@") == 15);
+	ret &= bfs_check(asciilen("@@@@@@@@@@@@@@@@\xFF@@@@@@@") == 16);
+	ret &= bfs_check(asciilen("@@@@@@@@@@@@@@@@a\xFF@@@@@@") == 17);
+	ret &= bfs_check(asciilen("@@@@@@@\xFF@@@@@@a\xFF@@@@@@@") == 7);
+	ret &= bfs_check(asciilen("@@@@@@@@\xFF@@@@@a\xFF@@@@@@@") == 8);
+	ret &= bfs_check(asciilen("@@@@@@@@@\xFF@@@@a\xFF@@@@@@@") == 9);
 
 	// From man 3p basename
 	ret &= check_base_dir("usr", ".", "usr");
@@ -68,6 +75,7 @@ bool check_bfstd(void) {
 	ret &= check_wordesc("\"word's\"", "'\"word'\\''s\"'", WESC_SHELL);
 	ret &= check_wordesc("\033[1mbold's\033[0m", "$'\\e[1mbold\\'s\\e[0m'", WESC_SHELL | WESC_TTY);
 	ret &= check_wordesc("\x7F", "$'\\x7F'", WESC_SHELL | WESC_TTY);
+	ret &= check_wordesc("~user", "\"~user\"", WESC_SHELL);
 
 	const char *charmap = nl_langinfo(CODESET);
 	if (strcmp(charmap, "UTF-8") == 0) {
@@ -75,7 +83,11 @@ bool check_bfstd(void) {
 		ret &= check_wordesc("\xF0\x9F", "$'\\xF0\\x9F'", WESC_SHELL | WESC_TTY);
 		ret &= check_wordesc("\xF0\x9F\x98", "$'\\xF0\\x9F\\x98'", WESC_SHELL | WESC_TTY);
 		ret &= check_wordesc("\xF0\x9F\x98\x80", "\xF0\x9F\x98\x80", WESC_SHELL | WESC_TTY);
+		ret &= check_wordesc("\xCB\x9Cuser", "\xCB\x9Cuser", WESC_SHELL);
 	}
 
+	ret &= bfs_check(xstrwidth("Hello world") == 11);
+	ret &= bfs_check(xstrwidth("Hello\1world") == 10);
+
 	return ret;
 }
