@@ -1,13 +1,15 @@
 // Copyright © Tavian Barnes <tavianator@tavianator.com>
 // SPDX-License-Identifier: 0BSD
 
+#include "prelude.h"
 #include "tests.h"
-#include "../src/ioq.h"
-#include "../src/bfstd.h"
-#include "../src/diag.h"
-#include "../src/dir.h"
+#include "ioq.h"
+#include "bfstd.h"
+#include "diag.h"
+#include "dir.h"
 #include <errno.h>
 #include <fcntl.h>
+#include <stdlib.h>
 
 /**
  * Test for blocking within ioq_slot_push().
@@ -38,15 +40,15 @@ static void check_ioq_push_block(void) {
 	const size_t depth = 2;
 
 	struct ioq *ioq = ioq_create(depth, 1);
-	bfs_verify(ioq, "ioq_create(): %s", xstrerror(errno));
+	bfs_everify(ioq, "ioq_create()");
 
 	// Push enough operations to fill the queue
 	for (size_t i = 0; i < depth; ++i) {
 		struct bfs_dir *dir = bfs_allocdir();
-		bfs_verify(dir, "bfs_allocdir(): %s", xstrerror(errno));
+		bfs_everify(dir, "bfs_allocdir()");
 
 		int ret = ioq_opendir(ioq, dir, AT_FDCWD, ".", 0, NULL);
-		bfs_verify(ret == 0, "ioq_opendir(): %s", xstrerror(errno));
+		bfs_everify(ret == 0, "ioq_opendir()");
 	}
 	bfs_verify(ioq_capacity(ioq) == 0);
 
