@@ -1,10 +1,10 @@
 // Copyright © Tavian Barnes <tavianator@tavianator.com>
 // SPDX-License-Identifier: 0BSD
 
+#include "prelude.h"
 #include "tests.h"
-#include "../src/trie.h"
-#include "../src/config.h"
-#include "../src/diag.h"
+#include "trie.h"
+#include "diag.h"
 #include <stdlib.h>
 #include <string.h>
 
