@@ -1,14 +1,14 @@
 // Copyright © Tavian Barnes <tavianator@tavianator.com>
 // SPDX-License-Identifier: 0BSD
 
+#include "prelude.h"
 #include "tests.h"
-#include "../src/xtime.h"
-#include "../src/bfstd.h"
-#include "../src/config.h"
+#include "xtime.h"
+#include "bfstd.h"
+#include "diag.h"
 #include <errno.h>
+#include <limits.h>
 #include <stdint.h>
-#include <stdio.h>
-#include <stdlib.h>
 #include <time.h>
 
 static bool tm_equal(const struct tm *tma, const struct tm *tmb) {
@@ -29,9 +29,9 @@ static bool check_one_xgetdate(const char *str, int error, time_t expected) {
 	int ret = xgetdate(str, &ts);
 
 	if (error) {
-		return bfs_check(ret == -1 && errno == error, "xgetdate('%s'): %s", str, xstrerror(errno));
+		return bfs_echeck(ret == -1 && errno == error, "xgetdate('%s')", str);
 	} else {
-		return bfs_check(ret == 0, "xgetdate('%s'): %s", str, xstrerror(errno))
+		return bfs_echeck(ret == 0, "xgetdate('%s')", str)
 			&& bfs_check(ts.tv_sec == expected && ts.tv_nsec == 0,
 				"xgetdate('%s'): %jd.%09jd != %jd",
 				str, (intmax_t)ts.tv_sec, (intmax_t)ts.tv_nsec, (intmax_t)expected);
@@ -81,13 +81,13 @@ static bool check_xgetdate(void) {
 /** Check one xmktime() result. */
 static bool check_one_xmktime(time_t expected) {
 	struct tm tm;
-	if (xlocaltime(&expected, &tm) != 0) {
-		bfs_diag("xlocaltime(%jd): %s", (intmax_t)expected, xstrerror(errno));
+	if (!localtime_r(&expected, &tm)) {
+		bfs_ediag("localtime_r(%jd)", (intmax_t)expected);
 		return false;
 	}
 
 	time_t actual;
-	return bfs_check(xmktime(&tm, &actual) == 0, "xmktime(" TM_FORMAT "): %s", TM_PRINTF(tm), xstrerror(errno))
+	return bfs_echeck(xmktime(&tm, &actual) == 0, "xmktime(" TM_FORMAT ")", TM_PRINTF(tm))
 		&& bfs_check(actual == expected, "xmktime(" TM_FORMAT "): %jd != %jd", TM_PRINTF(tm), (intmax_t)actual, (intmax_t)expected);
 }
 
@@ -99,6 +99,19 @@ static bool check_xmktime(void) {
 		ret &= check_one_xmktime(time);
 	}
 
+	// Attempt to trigger overflow (but don't test for it, since it's not mandatory)
+	struct tm tm = {
+		.tm_year = INT_MAX,
+		.tm_mon = INT_MAX,
+		.tm_mday = INT_MAX,
+		.tm_hour = INT_MAX,
+		.tm_min = INT_MAX,
+		.tm_sec = INT_MAX,
+		.tm_isdst = -1,
+	};
+	time_t time;
+	xmktime(&tm, &time);
+
 	return ret;
 }
 
@@ -124,6 +137,26 @@ static bool check_one_xtimegm(const struct tm *tm) {
 	return ret;
 }
 
+#if !BFS_HAS_TIMEGM
+/** Check an overflowing xtimegm() call. */
+static bool check_xtimegm_overflow(const struct tm *tm) {
+	struct tm copy = *tm;
+	time_t time = 123;
+
+	bool ret = true;
+	ret &= bfs_check(xtimegm(&copy, &time) == -1 && errno == EOVERFLOW);
+	ret &= bfs_check(tm_equal(&copy, tm));
+	ret &= bfs_check(time == 123);
+
+	if (!ret) {
+		bfs_diag("xtimegm(): " TM_FORMAT, TM_PRINTF(copy));
+		bfs_diag("(input):   " TM_FORMAT, TM_PRINTF(*tm));
+	}
+
+	return ret;
+}
+#endif
+
 /** xtimegm() tests. */
 static bool check_xtimegm(void) {
 	bool ret = true;
@@ -132,6 +165,7 @@ static bool check_xtimegm(void) {
 		.tm_isdst = -1,
 	};
 
+	// Check equivalence with mktime()
 	for (tm.tm_year =  10; tm.tm_year <= 200; tm.tm_year += 10)
 	for (tm.tm_mon  =  -3; tm.tm_mon  <=  15; tm.tm_mon  +=  3)
 	for (tm.tm_mday = -31; tm.tm_mday <=  61; tm.tm_mday +=  4)
@@ -141,16 +175,18 @@ static bool check_xtimegm(void) {
 		ret &= check_one_xtimegm(&tm);
 	}
 
+#if !BFS_HAS_TIMEGM
+	// Check integer overflow cases
+	ret &= check_xtimegm_overflow(&(struct tm) { .tm_sec = INT_MAX, .tm_min = INT_MAX });
+	ret &= check_xtimegm_overflow(&(struct tm) { .tm_min = INT_MAX, .tm_hour = INT_MAX });
+	ret &= check_xtimegm_overflow(&(struct tm) { .tm_hour = INT_MAX, .tm_mday = INT_MAX });
+	ret &= check_xtimegm_overflow(&(struct tm) { .tm_mon = INT_MAX, .tm_year = INT_MAX });
+#endif
+
 	return ret;
 }
 
 bool check_xtime(void) {
-	if (setenv("TZ", "UTC0", true) != 0) {
-		perror("setenv()");
-		return false;
-	}
-	tzset();
-
 	bool ret = true;
 	ret &= check_xgetdate();
 	ret &= check_xmktime();
